(*
defCFG - 
This module contains the definition of the Control Flow Graph datatype and the operations to work on it.
Authors: Pedro de Carvalho Gomes <pedrodcg@kth.se>
         Attilio Picoco <picoco@kth.se>
Date: 2013-02-02
*)

open Javalib_pack
open JCode
open Javalib
open JBasics
open JProgram
open JControlFlow
open JBir


(**
* Definition for the ControlFlowGraph
*)

type cfgFormat =
   | KTH
   | Normal

(**
* Mapping of the JBir.instr
*)
type edgeBirInstr =
   | Nop_Edge
   | AffectVar_Edge
   | AffectArray_Edge
   | AffectField_Edge
   | AffectStaticField_Edge
   | Goto_Edge
   | Ifd_Edge
   | Throw_Edge
   | Return_Edge
   | New_Edge
   | NewArray_Edge
   | InvokeStatic_Edge
   | InvokeVirtual_Edge
   | InvokeNonVirtual_Edge
   | MonitorEnter_Edge
   | MonitorExit_Edge
   | MayInit_Edge
   | Check_Edge

(**
* @param e edgeBirInstr BIR instruction
* @return int integer mapped
*
* It maps a BIR instruction edge to a number, for storing/loading purposes
*)
(* map_edge2int and map_int2edge should be modified coherently, if necessary *)
val map_edge2int : edgeBirInstr -> int

(**
* @param n int integer number
* @return edgeBirInstr BIR instruction mapped
*
* It maps a number to a BIR instruction edge, for storing/loading purposes
*)
(* map_edge2int and map_int2edge should be modified coherently, if necessary *)
val map_int2edge : int -> edgeBirInstr


(** a node can take as labels: ret = return, ent = entry, exp = exception *)
type t_node_label = string

(** an edge can take class names as labels *)
type t_edge_label = string

(** Nodes in a CFG - contains the information about
 * methods's full signature, the index in the instruction array
 * set of atomic propositions assigned to it and
 * a pair for unique identification from same nodes in same pc but with different semantics
 * *)
type t_cfg_node = {
   cmsig: JBasics.class_method_signature;
   pc: int;
   mutable nlabels: t_node_label list;
   (* Added excp and ret to differentiate them from the same node in same pc *)
   excp: bool;
   ret: bool;
}


(** Defines the labeling function *)
type t_cfg_trans = {
(*   inst: JBir.instr;*)
	inst: edgeBirInstr;
   mutable elabels: t_edge_label list;
}

(** Defines the edges as triples *)
type t_cfg_edge = (t_cfg_node * t_cfg_trans * t_cfg_node)

(**
* @param x Triple (a,_,_)
* @return first object a
*
* Returns the first element of a triple *)
val get_first : ('a * 'a * 'a) -> 'a

(** Define the CFG as a set of triples *)
module CFGSet : Set.S with type elt = t_cfg_edge

(** Define labels for entry and return nodes *)
val ret_lbl : string
val ent_lbl : string

(**
* @param n t_cfg_node
* @return string
*
* Generates a unique string *)
(* version with string *)
val get_node_id : t_cfg_node -> string
;;
(* version with hash number
val get_node_id : t_cfg_node -> string
 *)

(**
* @param tag_ret bool
* @param tag_excp bool
* @param tag_label t_node_label List
* @param pp JControlFlow.PP_Bir.t
* @param node_pc int
* @return t_cfg_node
*
* Function that creates nodes *)
val make_cfg_node : ?tag_ret:bool -> ?tag_excp:bool -> ?tag_label:(t_node_label list) -> JControlFlow.PP_Bir.t -> int -> t_cfg_node

(**
* @param tag_ret bool
* @param tag_excp bool
* @param tag_label t_node_label List
* @param pp JControlFlow.PP_Bir.t
* @param curr_meth_cms JBasics.class_method_signature
* @return t_cfg_node
*
* Function that creates nodes *)
val intra_make_cfg_node : ?tag_ret:bool -> ?tag_excp:bool -> ?tag_label:(t_node_label list) -> JControlFlow.PP_Bir.t -> JBasics.class_method_signature -> t_cfg_node

(**
* @param cms JBasics.class_method_signature
* @param pc int
* @param nl t_node_label List
* @param excp bool
* @param ret bool
* @return t_cfg_node
*
* Function that creates nodes *)
val make_cfg_node_pure : JBasics.class_method_signature -> int -> (t_node_label list) -> bool -> bool -> t_cfg_node

(**
* @param source_node t_cfg_node
* @param edge t_cfg_trans
* @param target_list t_cfg_node list
* @param cfg CFGSet.t
* @return CFGSet.t
*
* Add edges from source_node to all target_nodes in target_list *)
val add_cfg_edge : t_cfg_node -> t_cfg_trans -> t_cfg_node list -> CFGSet.t -> CFGSet.t

(**
* @param op JBir.instr
* @param ls t_edge_label list
* @return t_cfg_edge
*
* Function that creates edges *)
val make_cfg_edge : JBir.instr -> t_edge_label list -> t_cfg_trans

(**
* @param op edgeBirInstr
* @param ls t_edge_label list
* @return t_cfg_edge
*
* Function that creates edges *)
val make_cfg_edge_pure : edgeBirInstr -> t_edge_label list -> t_cfg_trans

(**
* @return t_cfg_node
*
* Function that creates a fake node
*)
val make_fake_node : t_cfg_node

(**
* @return t_cfg_edge
*
* Function that creates a fake edge
*)
val make_fake_edge : t_cfg_trans
