(*
openEnv - 
This module contains the definition of the Open Environment datatype and the operations to work on it.
It is used in order to perform the modular extraction of CFG from Java bytecode within the tool 'fgextractor'.
It is also related to the module jMCA.
Authors: Pedro de Carvalho Gomes <pedrodcg@kth.se>
         Attilio Picoco <picoco@kth.se>
Version: 1.0
Date: 2012-06-08
*)

open Javalib_pack
open JCode
open Javalib
open JBasics
open JProgram


(**
* Is the Open Environment supposed to include the classes not part of the program (but for example those from rt.jar)? YES
*)

(**
* INTERFACE ENVIRONMENT Data structure
*)
type interface_env =
{
	ext_interfaces: JBasics.class_name list; (* Interfaces extended *)
	methods: JBasics.class_method_signature list (* Methods provided *)
}

(**
* CLASS ENVIRONMENT Data structure
*)
type class_env =
{
	super: JBasics.class_name option; (* Class name of the superclass - option as JBasics.java_lang_object does not have a superclass *)
	impl_interfaces: JBasics.class_name list; (* Interfaces implemented *)
	fields: JBasics.field_signature list (* Fields of the class - it suffices, no need for class_field_signature *)
}

(**
* METHOD ENVIRONMENT Data structure
*)
type method_env =
{
	(* If the method is NOT provided, it is faked as concrete with empty body - that is, at least a return instruction -> its code is marked as NONE *)
	code: int option; (* Hash value of the instruction array *)
	handlers: JBasics.class_name list; (* Exceptions NOT propagated - it can be an empty list *)
	req_meths: JBasics.class_method_signature list (* Client methods required -> used only for missing methods AND not in the check for refinement *)
}

(**
* OPEN ENVIRONMENT Data structure
*)
type open_env =
{
	interface_map: interface_env ClassMap.t; (* Partial mapping of interface names *)
	class_map: class_env ClassMap.t; (* Partial mapping of class names *)
	method_map: method_env ClassMethodMap.t (* Partial mapping of method references *)
}

(**
* @param im => interface_map 
* @param cm => class_map
* @param mm => method_map
* @return open_env
*)
let create_open_env (im:interface_env ClassMap.t) (cm:class_env ClassMap.t) (mm:method_env ClassMethodMap.t) : open_env =
	{ interface_map = im; class_map = cm; method_map = mm }

(**
* @param ei => extended_interfaces
* @param m => provided methods
* @return interface_env
*)
let create_int_env (ei:JBasics.class_name list) (m:JBasics.class_method_signature list) : interface_env = 
	{ ext_interfaces = ei; methods = m }

(**
* @param s => super class 
* @param ii => implemented interfaces
* @param f => fields
* @return class_env
*)
let create_cl_env (s:JBasics.class_name option) (ii:JBasics.class_name list) (f:JBasics.field_signature list) : class_env =
	{ super = s; impl_interfaces = ii; fields = f }

(**
* @param c => hash value of the code
* @param h => handlers (exceptions not propagated)
* @return method_env
*)
let create_met_env (c:int option) (h:JBasics.class_name list) (rm:JBasics.class_method_signature list) : method_env =
	{ code = c; handlers = h; req_meths = rm }

(**
* @param old_values => class_name list
* @param new_values => class_name list
* @return bool
*
* old_values has to be a subset of new_values. This function is used by iar_interface
*)
let list_subset_cn (old_values:JBasics.class_name list) (new_values:JBasics.class_name list) : bool =
	(* old_values = subset of new_values *)
(*
	List.iter (fun e -> print_string ("OLD_CN: " ^ (JBasics.cn_name e) ^ "\n")) old_values;
	List.iter (fun e -> print_string ("NEW_CN: " ^ (JBasics.cn_name e) ^ "\n")) new_values;
	(List.for_all (fun e -> print_string ("CN: " ^ (JBasics.cn_name e) ^ "\n");(List.mem e new_values)) old_values)
*)	
	(List.for_all (fun e -> (List.mem e new_values)) old_values)

(**
* @param old_values => class_name list
* @param new_values => class_name list
* @return bool
*
* The two lists have to be the same. This function is used by both iar_interface and iar_class
*)
let list_eq_cn (old_values:JBasics.class_name list) (new_values:JBasics.class_name list) : bool =
	(* Not the same length: some elements are missing in either of the domains *)
	(* ALL the elements in a list have to be in the other *)
	(((List.length old_values) == (List.length new_values)) && (list_subset_cn old_values new_values))

(**
* @param old_values => class_method_signature list
* @param new_values => class_method_signature list
* @return bool
*
* old_values has to be a subset of new_values. This function is used by iar_method and by iar_interface
*)
let list_subset_cms (old_values:JBasics.class_method_signature list) (new_values:JBasics.class_method_signature list) : bool =
	(* old_values = subset of new_values *)
	(*
	List.iter (fun e -> print_string ("OLD_CMS: " ^ (JPrint.class_method_signature e) ^ "\n")) old_values;
	List.iter (fun e -> print_string ("NEW_CMS: " ^ (JPrint.class_method_signature e) ^ "\n")) new_values;
	(List.for_all (fun e -> print_string ("CMS: " ^ (JPrint.class_method_signature e) ^ "\n");(List.mem e new_values)) old_values)
*)	
	(List.for_all (fun e -> (List.mem e new_values)) old_values)

(**
* @param old_values => class_method_signature list
* @param new_values => class_method_signature list
* @return bool
*
* The two lists have to be the same. This function is used by iar_method and by iar_interface
*)
let list_eq_cms (old_values:JBasics.class_method_signature list) (new_values:JBasics.class_method_signature list) : bool =
	(* Not the same length: some elements are missing in either of the domains *)
	(* ALL the elements in a list have to be in the other *)
	(((List.length old_values) == (List.length new_values)) && (list_subset_cms old_values new_values))

(**
* @param old_values => field_signature list
* @param new_values => field_signature list
* @return bool
*
* old_values has to be a subset of new_values. This function is used by iar_class
*)
let list_subset_fs (old_values:JBasics.field_signature list) (new_values:JBasics.field_signature list) : bool =
	(* old_values = subset of new_values *)
	(*
	List.iter (fun e -> print_string ("OLD_FS: " ^ (JPrint.field_signature e) ^ "\n")) old_values;
	List.iter (fun e -> print_string ("NEW_FS: " ^ (JPrint.field_signature e) ^ "\n")) new_values;
	(List.for_all (fun e -> print_string ("FS: " ^ (JPrint.field_signature e) ^ "\n");(List.mem e new_values)) old_values)*)
	(List.for_all (fun e -> (List.mem e new_values)) old_values)

	

(**
* @param old_values => field_signature list
* @param new_values => field_signature list
* @return bool
*
* The two lists have to be the same. This function is used by iar_class
*)
let list_eq_fs (old_values:JBasics.field_signature list) (new_values:JBasics.field_signature list) : bool =
	(* Not the same length: some elements are missing in either of the domains *)
	(* ALL the elements in a list have to be in the other *)
	(((List.length old_values) == (List.length new_values)) && (list_subset_fs old_values new_values))

(**
* @param old_values => string list
* @param new_values => string list
* @return bool
*
* old_values has to be a subset of new_values. This function is used by iar_class and iar_method
* PATCH!!!
*)
let list_subset_str (old_values:string list) (new_values:string list) : bool =
	(List.for_all (fun e -> (List.mem e new_values)) old_values)

(**
* @param old_values => string list
* @param new_values => string list
* @return bool
*
* The two lists have to be the same. This function is used by iar_class and iar_method
* PATCH!!!
*)
let list_eq_str (old_values:string list) (new_values:string list) : bool =
	(* Not the same length: some elements are missing in either of the domains *)
	(* ALL the elements in a list have to be in the other *)
	(((List.length old_values) == (List.length new_values)) && (list_subset_str old_values new_values))
(**
* @param old_im => old Open Environment.interface_map
* @param new_im => new Open Environment.interface_map
* @return bool
*
* This function checks if the new Open Environment.interface_map is a refinement of the old one, following the rules given in the paper. 
*)
let iar_interface (old_im:interface_env ClassMap.t) (new_im:interface_env ClassMap.t) : bool = 
	let old_values = (ClassMap.key_elements old_im)
	and new_values = (ClassMap.key_elements new_im)
	in
	(* print_string ("\nINTERFACE\n"); *)
	(* The Domains have to be the same *)
	(list_eq_cn old_values new_values) && 
	(* The two maps are guaranteed to have the same elements and they exist *)
	(List.for_all (
		fun e -> 
			(*
			let _ = print_string ("INT: " ^ (JBasics.cn_name e) ^ "\n")
			in
			*)
			try
			let old_e = ClassMap.find e old_im
			and new_e = ClassMap.find e new_im
			in
			(* Check the methods - the cms lists have to be the same *)
			(
			
			(list_eq_cms old_e.methods new_e.methods) &&
			(* Check the interfaces - the new cn list has to be a subset of the old one *)
			(list_subset_cn new_e.ext_interfaces old_e.ext_interfaces)
			)
			with Not_found -> false (* There is something wrong in the maps *)
		) old_values)

(**
* @param old_cm => old Open Environment.class_map
* @param new_cm => new Open Environment.class_map
* @return bool
*
* This function checks if the new Open Environment.class_map is a refinement of the old one, following the rules given in the paper. 
*)
let iar_class (old_cm:class_env ClassMap.t) (new_cm:class_env ClassMap.t) : bool = 
	let old_values = (ClassMap.key_elements old_cm)
	and new_values = (ClassMap.key_elements new_cm)
	in
	(*print_string ("\nCLASS\n");*)
	(* The Domains have to be the same *)
	(list_eq_cn old_values new_values) &&
	(* The two maps are guaranteed to have the same elements and they exist *)
	(List.for_all (
		fun e ->
			(*
			let _  = print_string ("CL: " ^ (JBasics.cn_name e) ^ "\n")
			in
			*)
			try
			(*
			let old_e = ClassMap.find e old_cm
			and _ = print_string ("OLD_E found!\n")
			and new_e = ClassMap.find e new_cm
			and _ = print_string ("NEW_E found!\n")
			in
			*)
			let old_e = ClassMap.find e old_cm
			and new_e = ClassMap.find e new_cm
			in
			(* Check the superclass - The new one has to be the a subclass of the old one *)
			(* Only java.lang.Object has None as superclass *)
			(
			(match (old_e.super,new_e.super) with
				| None, None -> (JBasics.cn_equal e JBasics.java_lang_object)
					(* Just a further check *)
					(* true *) (* Both are supposed to be java.lang.Object *)
				| Some old_cn, Some new_cn -> 
					(* New one same as the old one *)
					(JBasics.cn_equal old_cn new_cn)
				| _,_ -> false (* One is java.lang.Object and the other is another class - something went wrong *)
			)
			&&
			(* Check the interfaces - The new cn list has to be a subset of the old one *)
			(list_subset_cn new_e.impl_interfaces old_e.impl_interfaces)
			&&
			(* Check the fields - The old fs list has to be a subset of the new one *)
			(list_subset_fs old_e.fields new_e.fields)
			)
			with Not_found -> false (* There is something wrong in the maps *)
	) old_values)

(**
* @param old_mm => old Open Environment.method_map
* @param new_mm => new Open Environment.method_map
* @param all_meths_excp_map => JBasics.class_name list Javalib_pack.JBasics.ClassMethodMap.t Map of propagated exceptions for each cms
* @return bool
*
* This function checks if the new Open Environment.method_map is a refinement of the old one, following the rules given in the paper. 
*)
let iar_method (old_mm:method_env ClassMethodMap.t) (new_mm:method_env ClassMethodMap.t) (all_meths_excp_map:(JBasics.class_name list) Javalib_pack.JBasics.ClassMethodMap.t) : bool =
	let old_values = (ClassMethodMap.key_elements old_mm)
	and new_values = (ClassMethodMap.key_elements new_mm)
	in
	(* print_string ("\nMETHOD\n"); *)
	(* The Domains have to be the same *)
	(list_eq_cms old_values new_values) &&
	(* The two maps are guaranteed to have the same elements and they exist *)
	(List.for_all (
		fun e -> 
			try
			(*let _ = print_string ("METH: " ^ (JPrint.class_method_signature e) ^ "\n") in*)
			(*
			let old_e = ClassMethodMap.find e old_mm
			and _ = print_string ("OLD_E found!\n")
			and new_e = ClassMethodMap.find e new_mm
			and _ = print_string ("NEW_E found!\n")
			in
			*)
			let old_e = ClassMethodMap.find e old_mm
			and new_e = ClassMethodMap.find e new_mm
			in
			match (old_e.code,new_e.code) with
			(* Both methods have empty bodies *)
			| None, None -> (list_subset_cn old_e.handlers new_e.handlers)
			(* Both methods are provided *)
			| Some old_c, Some new_c -> 
(*
				let _ = print_string ("OLD_C: " ^ (string_of_int old_c) ^ " - NEW_C: " ^ (string_of_int new_c) ^ "\n") in
				let b = (old_c == new_c && (list_eq_cn old_e.handlers new_e.handlers))
				in
				print_string ("SOME_SOME: " ^ (string_of_bool b) ^ "\n");b
*)

			(old_c == new_c && (list_eq_cn old_e.handlers new_e.handlers))
			(* The new method might be a refinement *)
			(* The new method cannot propagate any of the handled exceptions by the old one *)
			| None, Some _ -> 
				let prop_excp = 
					(try
					ClassMethodMap.find e all_meths_excp_map
					with Not_found -> [] (* No propagated exception *)
					)
				in
(*
				let b = (List.for_all (fun cms -> (List.mem cms prop_excp == false)) old_e.handlers)
				in
				print_string ("NONE_SOME: " ^ (string_of_bool b) ^ "\n");b
*)
				(List.for_all (fun cms -> (List.mem cms prop_excp == false)) old_e.handlers)
		(* It is not a refinement - I have not provided this method anymore in the new execution *)
			| Some _, None -> false 
			with Not_found -> false (* There is something wrong in the maps *)
	) old_values)


(*	(
	(List.iter (
		fun e -> 
			let _ = print_string ("METH: " ^ (JPrint.class_method_signature e) ^ "\n") in
			(*
			let old_e = ClassMethodMap.find e old_mm
			and _ = print_string ("OLD_E found!\n")
			and new_e = ClassMethodMap.find e new_mm
			and _ = print_string ("NEW_E found!\n")
			in
			*)
			let old_e = ClassMethodMap.find e old_mm
			and new_e = ClassMethodMap.find e new_mm
			in
let b =
			match (old_e.code,new_e.code) with
			(* Both methods have empty bodies *)
			| None, None -> (list_subset_cn old_e.handlers new_e.handlers)
			(* Both methods are provided *)
			| Some old_c, Some new_c -> 

				let _ = print_string ("OLD_C: " ^ (string_of_int old_c) ^ " - NEW_C: " ^ (string_of_int new_c) ^ "\n") in
				let b = (old_c == new_c && (list_eq_cn old_e.handlers new_e.handlers))
				in
				print_string ("SOME_SOME: " ^ (string_of_bool b) ^ "\n");b

			(*(old_c == new_c && (list_eq_cn old_e.handlers new_e.handlers))*)
			(* The new method might be a refinement *)
			(* The new method cannot propagate any of the handled exceptions by the old one *)
			| None, Some _ -> (list_subset_cn old_e.handlers new_e.handlers)
		(* It is not a refinement - I have not provided this method anymore in the new execution *)
			| Some _, None -> false 

		in print_string ""
	) old_values)
	; false)
*)


(**
* @param old_oe => old Open Environment
* @param new_oe => new Open Environment
* @param all_meths_excp_map => JBasics.class_name list Javalib_pack.JBasics.ClassMethodMap.t Map of propagated exceptions for each cms
* @return bool
*
* This function checks if the new Open Environment is a refinement of the old one, following the rules given in the paper. 
*)
let is_a_refinement (old_oe:open_env) (new_oe:open_env) (all_meths_excp_map:(JBasics.class_name list) Javalib_pack.JBasics.ClassMethodMap.t) : bool =
		
	((iar_interface old_oe.interface_map new_oe.interface_map) && (iar_class old_oe.class_map new_oe.class_map) && (iar_method old_oe.method_map new_oe.method_map all_meths_excp_map))
	(*
	let b1 = (iar_interface old_oe.interface_map new_oe.interface_map)
	and b2 = (iar_class old_oe.class_map new_oe.class_map)
	and b3 = (iar_method old_oe.method_map new_oe.method_map all_meths_excp_map)
	in
	print_string ("INTERFACE: " ^ (string_of_bool b1) ^ " - CLASS: " ^ (string_of_bool b2) ^ " - METHOD: " ^ (string_of_bool b3) ^ "\n");
	(b1 && b2 && b3)
	*)

(**
* @param cms => class_method_signature to check if provided or not
* @param oe => open_env
* @return bool
*
* This checks if the method is provided or not
*)
let is_provided (cms:JBasics.class_method_signature) (oe:open_env) : bool =
	try
	let e = ClassMethodMap.find cms oe.method_map
	in
	(match e.code with | None -> false (* It is annotated *)| Some _ -> true )
	with _ -> false (* The method has not been parsed properly in the previous steps of the execution or the OpenEnv considered is an empty one *)

(**
* @return open_env
*
* This creates an empty OpenEnv
*)
let empty_open_env : open_env =
	{
		interface_map = (ClassMap.empty:interface_env ClassMap.t);
		class_map = (ClassMap.empty:class_env ClassMap.t);
		method_map = (ClassMethodMap.empty:method_env ClassMethodMap.t)
	}

(**
* @param oe => open_env
* @param cms => JBasics.class_method_signature
* @return JBasics.class_name list
*)
let get_handlers (oe:open_env) (cms:JBasics.class_method_signature) : (JBasics.class_name list) =
	try
	(ClassMethodMap.find cms oe.method_map).handlers
	with Not_found -> []
(**
* Name of the annotation to check against a method and its element
*)
let annotation_cn = "GhostComponent" (* Annotation-type *)
let annotation_elem_name_hndlr = "handlers"
let annotation_elem_name_rqmts = "req_meths"

exception Bad_Annotation of string

let name = "[a-zA-Z0-9$_]+"
let blank_space = "[ \t]+"
let dot = "[.]"

(**
* Checking if the string is well-formed.
* The string is made as "RETURN_TYPE FULL_PACKAGE_CLASS_NAME.METHOD_NAME(ARG1,...,ARGN)"
*)
let check_cms_well_formed (s:string) : bool =
	let package_cn = "\\(" ^ name ^ "\\(" ^ dot ^ name ^ "\\)*\\)" 
	(* Including no packages and primitives *)
	and met_name = "\\(<init>\\|" ^ name ^ "\\)"
	in
	let arg_r = package_cn ^ "\\(\\[\\]\\)*"
	in
	(* brackets needed *)
	let arg_list_r = "(\\(" ^ arg_r ^ "\\(," ^ arg_r ^ "\\)*\\)*)"
	in
	let cms_r = Str.regexp (arg_r ^ blank_space ^ package_cn ^ dot ^ met_name ^ arg_list_r)
	in
	if (String.compare s "") == 0 || (Str.string_match cms_r s 0) == false
	then false
	else true
	
(**
* Parsing a cms from a string
* The string is made as "RETURN_TYPE FULL_PACKAGE_CLASS_NAME.METHOD_NAME(ARG1,...,ARGN)"
*)
let parse_cms (s:string) : JBasics.class_method_signature =
	if check_cms_well_formed s == false
	then raise (Bad_Annotation annotation_elem_name_rqmts)
	else
	begin
	let build_cn fp_cn_l = 
		let first = List.hd fp_cn_l
		and tail = List.tl fp_cn_l
		in
		(JBasics.make_cn (List.fold_left (fun so_far piece -> (so_far ^ "." ^ piece)) first tail))
	in
	let get_lvl_bt_s arg_s = 
		let n_pos_array = 
		begin
		try
		Str.search_forward (Str.regexp ("\\[\\]")) arg_s 0
		with Not_found -> -1
		end
		in
		begin
		if n_pos_array == -1
		then (* No array *)
		(0,arg_s)
		else
	((((String.length arg_s) - n_pos_array)/2),(Str.string_before arg_s n_pos_array))
		end
	and arg_type basic_type_s n_levels = 
		begin
		let rec level_down elem lvl =
			begin
			if lvl == 0 then elem
			else
			(JBasics.TObject (JBasics.TArray (level_down elem (lvl - 1))))
			end
		and basic_type =
			match basic_type_s with
			| "boolean" -> (JBasics.TBasic (`Bool)) (* Java keyword, not Sawja *)
			| "float" -> (JBasics.TBasic (`Float))
			| "int" -> (JBasics.TBasic (`Int))
			| "long" -> (JBasics.TBasic (`Long))
			| "short" -> (JBasics.TBasic (`Short))
			| "byte" -> (JBasics.TBasic (`Byte))
			| "double" -> (JBasics.TBasic (`Double))
			| "char" -> (JBasics.TBasic (`Char))
			| _ -> (JBasics.TObject (JBasics.TClass (JBasics.make_cn basic_type_s)))
		in 
		(level_down basic_type n_levels)
		end

	in
	let build_args args_list_s =
		if (String.compare args_list_s "") == 0
		then []
		else
		begin
		let args_s = Str.split (Str.regexp ",") args_list_s
		in
		List.rev (List.fold_left (fun l arg_s -> 
						let (n_lvl,bt_s) = (get_lvl_bt_s arg_s)
						in
						(arg_type bt_s n_lvl)::l
			)[] args_s)
		end
	and build_ret ret_s =
		let (n_lvl,bt_s) = (get_lvl_bt_s ret_s)
		in
		if (String.compare bt_s "void") == 0 then None
		else Some (arg_type bt_s n_lvl)
	in
	let s_parts = Str.split (Str.regexp blank_space) s
	in
	let cn_met_name_args_list = Str.split (Str.regexp "(") (List.hd (List.tl s_parts))
	and return_type_s = List.hd s_parts
	in
	let full_package_cn_met_name_list =
		List.rev (Str.split (Str.regexp dot) (List.hd cn_met_name_args_list))
	and args_l = (List.hd (List.tl cn_met_name_args_list))
	in
	let full_package_cn_list = List.rev (List.tl full_package_cn_met_name_list)
	and met_name = List.hd full_package_cn_met_name_list
	(* removing the close bracket *)
	and args_list_s = Str.string_before args_l ((String.length args_l) -1)
	in
	let cn = build_cn full_package_cn_list
	and args = build_args args_list_s
	and ret = build_ret return_type_s
	in
	JBasics.make_cms cn (JBasics.make_ms met_name args ret)
	end

(*
(**
* Parsing a string representing a class_method_signature for getting the class_name.
* The string is made as "RETURN_TYPE FULL_PACKAGE_CLASS_NAME.METHOD_NAME(ARG1,...,ARGN)"
*)
let get_cn_from_cms (s:string) : (JBasics.class_name) =
	if check_cms_well_formed s == false
	then raise (Bad_Annotation annotation_elem_name_rqmts)
	else
	begin
	let s_parts = Str.split (Str.regexp blank_space) s
	in
	let cn_met_name_args_list = Str.split (Str.regexp "(") (List.hd (List.tl s_parts))
	in
	let full_package_cn_met_name_list = Str.split (Str.regexp dot) (List.hd cn_met_name_args_list)
	in
	let full_package_cn_list = List.rev (List.tl (List.rev full_package_cn_met_name_list))
	in
	let first = List.hd full_package_cn_list
	and tail = List.tl full_package_cn_list
	in
	let s = (List.fold_left (fun so_far piece -> (so_far ^ "." ^ piece)) first tail)
	in
	(JBasics.make_cn s)
	end
*)
(**
* Getting the list of class_method_signature referenced in req_meths annotation, from a set of annotations
*)
let get_cms_from_rm_ann  (m_anns:Javalib.method_annotations) : (JBasics.class_method_signature list) =
	try
	let annotation = Pervasives.fst (List.find (fun (ann,_) -> (String.compare (JBasics.cn_simple_name ann.kind) annotation_cn) == 0) m_anns.ma_global)
	in
	let cms_l =
	(try
	match Pervasives.snd (List.find (fun (s,_) -> (String.compare s annotation_elem_name_rqmts) == 0) annotation.element_value_pairs) with
	| JBasics.EVArray l -> 
		List.map (fun e -> match e with 
				| JBasics.EVCstString s -> parse_cms s
				| _ -> raise (Bad_Annotation annotation_elem_name_rqmts)
		) l
	| _ -> raise (Bad_Annotation annotation_elem_name_rqmts)
(* This should not happen, if the class files have been compiled correctly => the annotation contains necessarily the field 'req_meths' *)
	with Not_found -> raise (Bad_Annotation annotation_elem_name_rqmts) (* annotation missing *)
	)
	in cms_l
	with 
	| Bad_Annotation ann_field -> failwith ("ANNOTATION '" ^ annotation_cn ^ "' not used properly. Field '" ^ ann_field ^ "' is missing or not used correctly")
	| Not_found -> [] (* Not annotated with ours or at all *)


(**
* Getting the list of children exception given a parent node. The list is made of class_names. It checks if the name is already in the list.
*)
let get_parent_children_excp (parsed_prog:Javalib_pack.JCode.jcode JProgram.program) (excp_list:JBasics.class_name list) (parent_cn:JBasics.class_name) : (JBasics.class_name list) =
	begin
	try
	let parent_node = JProgram.get_node parsed_prog parent_cn
	in
	match parent_node with
	| JProgram.Interface _ -> excp_list
	| JProgram.Class c_parent_node ->
		begin
		let children_nodes = (JProgram.get_all_children_classes c_parent_node)
		in
		let children_cn =
		List.fold_left
		(fun l n ->
			let cn = JProgram.get_name (JProgram.Class n)
			in
			if List.mem cn excp_list == true
			then l
			else (cn::l)
		) [] children_nodes
		in
		(children_cn@excp_list)
		end
	with Not_found -> excp_list
	end 
(**
* @param parsed_prog => 'a JProgram.program
* @param ref_classes => JBasics.class_name list
* @param ref_meths => JBasics.class_method_signature list
* @return OpenEnv.open_env
*
* It creates an Open Environment structure for the parsed program - only classes and methods of the client
*)
let get_open_env (parsed_prog:Javalib_pack.JCode.jcode JProgram.program) (ref_classes:JBasics.class_name list) (ref_meths:JBasics.class_method_signature list) : open_env =

	let (int_m,cl_m) = 
		(ClassMap.fold
		(fun cn node (int_m,cl_m) ->
			if List.mem cn ref_classes == true
			(* In the client code *)
			then
			begin
			match node with
			| JProgram.Class cnode ->
				begin
				let fields_list = (FieldMap.key_elements (JProgram.get_fields node))
				and super_class_name = 
					(match (JProgram.super_class node) with
					(* java.lang.Object is supposed to be the only one with NONE *)
					| None -> None
					| Some sc ->
						(Some (JProgram.get_name (JProgram.Class (sc))))
					)
(*
(* Check if java_lang_object has super_class = NONE => YES *)
in
let _ = print_string ("ClassName: " ^ (JBasics.cn_name cn) ^ " - Super: " ^ (match super_class_name with | Some c -> (JBasics.cn_name c) | None -> "NONE") ^ "\n")
*)
				and imp_int = 				
				(List.fold_left (fun l i -> (JProgram.get_name (JProgram.Interface (i)))::l) [] (JProgram.implemented_interfaces cnode))
				in 
				let c_env = (create_cl_env super_class_name imp_int fields_list)
				in
				(int_m,(ClassMap.add cn c_env cl_m))
				end
			| JProgram.Interface inode ->
				begin
				let sup_int = 
				(List.fold_left (fun l i -> (JProgram.get_name (JProgram.Interface (i)))::l) [] (JProgram.super_interfaces inode))
				and met_prov = 
					(List.map
					(fun e -> (Javalib.get_class_method_signature e) )
					(* node is an interface_node => getting only abstract methods *)
					(MethodMap.value_elements (JProgram.get_methods node)))
				in
				let i_env = (create_int_env sup_int met_prov)
				in
				((ClassMap.add cn i_env int_m),cl_m)
				end
			end
			else
			(int_m,cl_m)
		) 
		parsed_prog.classes
		((ClassMap.empty:interface_env ClassMap.t),(ClassMap.empty:class_env ClassMap.t)))

	and met_m =
		(ClassMap.fold
		( fun cn node met_m ->
			if List.mem cn ref_classes == true
			then
			(* In the client code *)
			begin
			(* Adding to the map only the CONCRETE methods  - actual and fake *)
			match node with
			| JProgram.Interface _ -> met_m (* an interface_node provides only abstract methods *)
			| JProgram.Class _ ->
				let methods = (JProgram.get_concrete_methods node)
				in
				(MethodMap.fold
				( fun _ cm met_m ->
					if List.mem cm.cm_class_method_signature ref_meths == true
					then
					(* In the client code *)
					begin
					(* Check if it's ANNOTATED or ACTUALLY IMPLEMENTED *)
(* Don't care about visibility -> It is supposed to be set as RTInvisible as @Retention for @GhostComponent is set to RetentionPolicy.CLASS => we need it only in the class file *)
					try
					let annotation = Pervasives.fst (List.find (fun (ann,_) -> (String.compare (JBasics.cn_simple_name ann.kind) annotation_cn) == 0) cm.cm_annotations.ma_global)
					in
					(* Annotated - marking its code as NONE and its handlers as those declared in it *)
					let c = None
					and hndlrs =
						begin
						let parent_hndlrs = 
						(try
						match Pervasives.snd (List.find (fun (s,_) -> (String.compare s annotation_elem_name_hndlr) == 0) annotation.element_value_pairs) with
						| JBasics.EVArray l -> 
							List.map (fun e -> match e with 
									| JBasics.EVCstString s -> (JBasics.make_cn s)
									| _ -> raise (Bad_Annotation annotation_elem_name_hndlr)
							) l
						| _ -> raise (Bad_Annotation annotation_elem_name_hndlr)
	(* This should not happen, if the class files have been compiled correctly => the annotation contains necessarily the field 'handlers' *)
						
						with Not_found -> raise (Bad_Annotation annotation_elem_name_hndlr) (* annotation field missing *)
						)
						in
						let all_excp = parent_hndlrs
						in
						(* Getting the children *)
						(List.fold_left (fun l pcn -> get_parent_children_excp parsed_prog l pcn) all_excp parent_hndlrs)
						end
					and rq_mts =
						begin
						let parent_cms_l =
						(try
						match Pervasives.snd (List.find (fun (s,_) -> (String.compare s annotation_elem_name_rqmts) == 0) annotation.element_value_pairs) with
						| JBasics.EVArray l -> 
							List.map (fun e -> match e with 
									| JBasics.EVCstString s -> parse_cms s
									| _ -> raise (Bad_Annotation annotation_elem_name_rqmts)
							) l
						| _ -> raise (Bad_Annotation annotation_elem_name_rqmts)
	(* This should not happen, if the class files have been compiled correctly => the annotation contains necessarily the field 'req_meths' *)
						
						with Not_found -> raise (Bad_Annotation annotation_elem_name_rqmts) (* annotation field missing *)
						)
						in
						(* Getting the children *)
						List.fold_left (fun l p_cms ->
							let (p_cn,p_ms) = JBasics.cms_split p_cms
							in
							try
							let children_nodes =
								match JBasics.ms_name p_ms with
								(* Children have their own constructor *)
								| "<init>" -> []
								| _ ->
								begin
								match JProgram.get_node parsed_prog p_cn with
								|JProgram.Class p_node -> (JProgram.get_all_children_classes p_node)
								|JProgram.Interface _ -> []
								end
							in
							let subhierarchy_l = 
							begin
							let children_cn =
							List.map (fun c_node ->
								JProgram.get_name (JProgram.Class (c_node))
						) children_nodes
							in
							let children_cms =
							List.map (fun c_cn ->
								JBasics.make_cms c_cn p_ms) children_cn
							in
							p_cms::children_cms
							end
							in
							subhierarchy_l@l
							(* it should not happen *)
							with Not_found -> failwith ("Class " ^ (JBasics.cn_name p_cn) ^ " not found. Leaving")
						) [] parent_cms_l
						end
					in
					let m_env = (create_met_env c hndlrs rq_mts)
					in
					(ClassMethodMap.add cm.cm_class_method_signature m_env met_m)
					with 
					| Bad_Annotation ann_field -> failwith ("ANNOTATION '" ^ annotation_cn ^ "' not used properly for method " ^ (JPrint.class_method_signature cm.cm_class_method_signature) ^ ". Field '" ^ ann_field ^ "' is missing or not used correctly")
					| Not_found ->
					(* Actually implemented *)
					match cm.cm_implementation with
					| Javalib.Java j ->
						begin
						let jcode = (Lazy.force j)
						in
						let c = Some (Hashtbl.hash jcode.JCode.c_code)
						and hndlrs =
							begin
							let parent_hndlrs = 
							(List.map 
							(fun eh -> 
							match eh.e_catch_type with
							(* what to do with NONE match? *)
							| None -> (JBasics.make_cn "java.lang.Throwable")
							| Some ect -> ect
							) 
							jcode.JCode.c_exc_tbl)
							in
							let all_excp = parent_hndlrs
							in
							(* Getting the children *)
							(List.fold_left (fun l pcn -> get_parent_children_excp parsed_prog l pcn) all_excp parent_hndlrs)
							end
						and rq_mts = []
						in
						let m_env = (create_met_env c hndlrs rq_mts)
						in
						(ClassMethodMap.add cm.cm_class_method_signature m_env met_m)
						end
					|Javalib.Native -> met_m
					end
					else
					met_m
				)
				methods
				met_m)
			end
			else
			met_m				
				)
		parsed_prog.classes
		(ClassMethodMap.empty:method_env ClassMethodMap.t)
		)
		in
		(create_open_env int_m cl_m met_m)


(**
* @param old_oe => OpenEnv.open_env
* @param new_oe => OpenEnv.open_env
* @param cms => JBasics.class_method_signature
* @return bool
*
* It checks if a method has changed its code => either newly provided or removed or replaced
*)
let has_changed (old_oe:open_env) (new_oe:open_env) (cms:JBasics.class_method_signature) : bool =
(* The cms is supposed to be at least in the old_oe, as this function is called on the old CFG referring to the old_oe - If called on the methods having a CFG extracted, it is expected that the given cms had code in the old environment  => the case old_e.code == None is not supposed to happen *)
	try
	let old_e = ClassMethodMap.find cms old_oe.method_map
	and new_e = ClassMethodMap.find cms new_oe.method_map
	in
	match (old_e.code,new_e.code) with
	(* Not provided yet *)
	| None, None -> Pervasives.not (list_subset_cn old_e.handlers new_e.handlers)
	(* Might have been replaced *)
	| Some old_c, Some new_c -> Pervasives.not ((old_c == new_c) && (list_subset_cn old_e.handlers new_e.handlers))
	(* newly provided *)
	| None, Some _ -> true
	(* removed *)
	| Some _, None -> true

	with Not_found -> true (* Expected not to be included in the new_oe *)

(**
*@param oe => OpenEnv.open_env
*@return bool
*
* It checks if an Open Environment is properly defined or if it's empty
*)
let is_empty (oe:open_env): bool =
	(
	(ClassMap.is_empty oe.interface_map)
	&&
	(ClassMap.is_empty oe.class_map)
	&&
	(ClassMethodMap.is_empty oe.method_map)
	)

(**
*@param old_oe => OpenEnv.open_env
*@param invocation_map => JBasics.class_method_signature list JBasics.ClassMethodMap.t
*@return bool
*)
let check_rm_anns (old_oe:open_env) (invocation_map:(JBasics.class_method_signature list) ClassMethodMap.t) : bool =
(* Each cms has its CFG either newly computed or recomputed *)
	ClassMethodMap.fold (fun cms inv_l so_far ->
		let b = 
			if is_provided cms old_oe == false
			then
			begin
			try
			let old_e = ClassMethodMap.find cms old_oe.method_map
			in
			list_subset_cms inv_l old_e.req_meths
		(* Method newly added - the first check is misleading in this context - not considered *)
			with Not_found -> true
			end
			else true (* Method already provided -> maybe CFG recomputed for some changes *)
		in
		if b == false
		then print_string ("Method " ^ (JPrint.class_method_signature cms) ^ " does NOT respect its own previous annotation about the required methods\n");
		(b && so_far)
		) invocation_map true
